class RelationDescriptor:
    def __init__(self, *trackings):
        from rogue.models.fields import RelationField

        self._trackings = trackings
        self._formatted_trackings = []

        for tracking in self._trackings:
            if isinstance(tracking, RelationField):
                self._formatted_trackings.append(
                    {
                        "left_table_name": tracking._parent.table_name,
                        "left_field_name": tracking.name,
                        "right_table_name": tracking._foreign_model.table_name,
                        "right_field_name": "id",
                    }
                )

    def __iter__(self):
        return iter(self._formatted_trackings)



from rogue.backends.sqlite.client import DatabaseClient
from rogue.backends.sqlite.schema import DatabaseSchemaEditor, DatabaseSchemaReader
from rogue.models.utils import get_all_models
from rogue.settings import settings


def makemigrations(db_name=None):
    db_name = db_name or settings.DATABASE_NAME

    migrator = MigrationCreator(db_name=db_name)
    migration_name = migrator.create_migration()

    print(f"Migration {migration_name} created!")


def migrate(filename, db_name=None):
    if filename is None:
        raise ValueError(
            "A value must be passed for filename when running the migrate command."
        )

    db_name = db_name or settings.DATABASE_NAME
    db_client = DatabaseClient(db_name)

    with open(filename) as file:
        for statement in file.read().split(";"):
            print(statement)
            db_client.execute(statement)
            print("Successful!")


class MigrationCreator:
    def __init__(self, db_name) -> None:
        self.db_name = db_name
        self._db_client = DatabaseClient(self.db_name)
        self._schema_reader = DatabaseSchemaReader(self._db_client)
        self._schema_editor = DatabaseSchemaEditor(self._db_client)

    def create_migration(self):
        diff = self.process_differences()
        sql = self._schema_editor.build_sql_migration(diff)
        migration_name = self.get_migration_name()

        with open(migration_name, "w") as file:
            file.write(sql)

        return migration_name

    def get_migration_name(self):
        # TODO: Add a way to keep track of migration names
        return "migration.sql"

    def process_differences(self):
        current_database_schema = self.analyze_database()
        new_database_schema = self.models_schema()

        tables_to_create = {}
        tables_to_modify = {}
        tables_to_delete = []

        for table_name, rows in new_database_schema.items():
            if table_name not in current_database_schema:
                tables_to_create[table_name] = rows
                continue

            if rows != current_database_schema[table_name]:
                changes = self._get_rows_diff(current_database_schema[table_name], rows)
                tables_to_modify[table_name] = changes

        for table_name in current_database_schema:
            if table_name not in new_database_schema:
                tables_to_delete.append(table_name)

        return {
            "create": tables_to_create,
            "alter": tables_to_modify,
            "delete": tables_to_delete,
        }

    def analyze_database(self):
        schema = {}

        tables = self._schema_reader.get_tables()
        for table in tables:
            schema[table] = self._schema_reader.get_rows(table)

        return schema

    def models_schema(self):
        schema = {}

        all_models = get_all_models()
        for model in all_models:
            rows = []
            for row in model.get_model_fields().values():
                # TODO: Make smarter, and based on the database backend used
                row_type = "integer" if row.PYTHON_TYPE is int else "text"
                rows.append(
                    {
                        "name": row.name,
                        "type": row_type,
                        "notnull": not row.nullable,
                        "default_value": row.default,
                        "is_pk": row.is_pk,
                    }
                )

            schema[model.table_name] = rows

        return schema

    def _get_rows_diff(self, old_rows, new_rows):
        create = []
        alter = []
        delete = []
        no_change = []

        old_rows = {row["name"]: row for row in old_rows}
        new_rows = {row["name"]: row for row in new_rows}

        for name in old_rows:
            if name not in new_rows:
                delete.append(name)

        for name, row in new_rows.items():
            if name not in old_rows:
                create.append(row)
                continue

            if row != old_rows[name]:
                alter.append(row)
            else:
                no_change.append(row)

        return {
            "create": create,
            "alter": alter,
            "delete": delete,
            "no_change": no_change,
        }



from ..base import BaseQueryBuilder
from ..errors import OperationalError


class QueryBuilder(BaseQueryBuilder):
    def fetch_one(self):
        data = self.client.execute(self._build_select()).fetchone()
        return self._format_output_data([data])

    def fetch_all(self):
        data = self.client.execute(self._build_select()).fetchall()
        return self._format_output_data(data)

    def insert(self, data):
        self.client.execute(*self._build_insert(data))
        data = (
            self.__class__(self.client, self.model)
            .where(
                table_name=self.model.table_name,
                field="id",
                comparison=self.EQUAL,
                value="(SELECT last_insert_rowid())",
            )
            .fetch_one()
        )
        return data

    def update(self, pk, data):
        self.client.execute(*self._build_update(pk, data))
        data = (
            self.__class__(self.client, self.model)
            .where(
                table_name=self.model.table_name,
                field="id",
                comparison=self.EQUAL,
                value=pk,
            )
            .fetch_one()
        )
        return data

    def delete(self):
        self.client.execute(*self._build_delete())

    def _format_fields(self):
        fields = []

        for field in self.fields:
            fields.append(".".join((self.table_name, field)))

        return fields

    def _build_select(self):
        query = f"{self.SELECT} {', '.join(self._format_fields())} {self.FROM} {self.table_name}"

        if self.where_statements:
            query = f"{query} {self._build_where()}"

        return query

    def _build_insert(self, data):
        assert not self.where_statements, "No where can be passed to an insert backend."

        if not data:
            return f"{self.INSERT} {self.table_name} DEFAULT {self.VALUES}", ()

        self._validate_data(data)
        headers, formatted_data = self._format_input_data(data)

        return (
            f"{self.INSERT} {self.table_name} ({', '.join(headers)}) {self.VALUES} ("
            f"{', '.join(['?' for _ in range(len(headers))])})",
            formatted_data,
        )

    def _build_update(self, pk, data):
        self._validate_data(data)
        headers, formatted_data = self._format_input_data(data)

        formatted_column_updates = [f"{col_name} = ?" for col_name in headers]

        return (
            f"{self.UPDATE} {self.table_name} SET {', '.join(formatted_column_updates)}",
            formatted_data,
        )

    def _build_delete(self):
        query = f"{self.DELETE} {self.FROM} {self.table_name}"

        if self.where_statements:
            query = f"{query} {self._build_where()}"

        return query, ()

    def _build_where(self):
        wheres = []
        joins = []
        for where in self.where_statements:
            if where.relation_descriptor:
                for relation in where.relation_descriptor:
                    joins.append(
                        f"{self.INNER_JOIN} {relation['right_table_name']} {self.ON} "
                        f"{relation['left_table_name']}.{relation['left_field_name']} = "
                        f"{relation['right_table_name']}.{relation['right_field_name']}"
                    )

            wheres.append(
                f"{where.table_name}.{where.field} {where.comparison} {where.value}"
            )

        return f"{' '.join(joins)} {self.WHERE} {f' {self.AND} '.join(wheres)}"



import os
from importlib import import_module
from dotenv import load_dotenv

from rogue.settings import default_settings


load_dotenv()


SETTINGS_FILE_VAR_NAME = "ROGUE_ORM_SETTINGS"
DEFAULT_SETTINGS_FILE_NAME = "settings"


def _get_settings():
    settings_file = os.getenv(SETTINGS_FILE_VAR_NAME, DEFAULT_SETTINGS_FILE_NAME)

    try:
        settings_module = import_module(settings_file)
    except ImportError:
        settings_module = default_settings

        # For now, use the default settings if no setting was set in
        # environment variables. Eventually, we might want to enforce
        # the use of a settings file.
        # raise ValueError(
        #     f"{SETTINGS_FILE_VAR_NAME} was not properly set in "
        #     "the environment variables. This is required to use RogueORM."
        # )

    for setting in dir(default_settings):
        if setting.isupper() and not hasattr(settings_module, setting):
            setattr(settings_module, setting, getattr(default_settings, setting))

    return settings_module


settings = _get_settings()



import argparse

from rogue.migrations import makemigrations, migrate


actions = {
    "makemigrations": makemigrations,
    "migrate": migrate,
}


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Actions for the Rogue ORM.")
    parser.add_argument(
        "action",
        type=str,
        choices=actions.keys(),
        help="The action expected from Rogue.",
    )
    parser.add_argument(
        "--db_name",
        type=str,
        help="The name of the database to use. Will default to the settings database if not specified.",
    )
    parser.add_argument(
        "--filename",
        type=str,
        help="The name of the file to use to migrate. Eventually, this will be deprecated in favor of a better migration tracking solution.",
    )

    args = parser.parse_args()

    print(args)

    kwargs = {"db_name": args.db_name}

    if args.action == "migrate":
        kwargs["filename"] = args.filename

    actions[args.action](**kwargs)



from rogue.settings import settings

from ..base import BaseDatabaseClient
import sqlite3

from ..errors import OperationalError


class DatabaseClient(BaseDatabaseClient):
    def get_connection(self):
        if self._connection is None:
            self._connection = sqlite3.connect(self._db_name)

        return self._connection

    def execute(self, statement, args=()):
        connection = self.get_connection()
        cursor = connection.cursor()
        data = cursor.execute(statement, args)
        connection.commit()

        return data

    def close(self):
        self.get_connection().close()



from collections.abc import Iterable
from typing import get_args

from rogue.managers import RelationManager, ManyToManyManager
from rogue.query import InLookup, Lookup

from .errors import FieldValidationError
from .utils import get_through_model


UNKNOWN_VALUE = "__UNKNOWN_VALUE__"

# Necessary so that we can reference Field in FieldMeta
Field = None


class FieldMeta(type):
    def __getitem__(self, item: type):
        if isinstance(item, tuple):
            assert len(item) == 1, "Only 1 item can be passed in the Field typing."
            item = item[0]

        args = get_args(item)
        nullable = type(None) in args
        assert (
            len(args) < 2 or len(args) == 2 and nullable
        ), "Only union between a type and None is accepted."

        if args:
            type_ = [arg for arg in args if arg is not None][0]
        else:
            type_ = item

        return Field(type_, nullable=nullable)


class Field(metaclass=FieldMeta):
    def __init__(self, type_: type, nullable: bool = False):
        from .base import Model

        self.type: type = type_
        self.nullable: bool = nullable

        self._kwargs = {}

        if type_ not in FIELD_MAPPING and not issubclass(type_, Model):
            raise TypeError(f"{type_} is not supported.")

        if issubclass(type_, Model):
            self._field = ForeignKeyField
            self._kwargs["foreign_model_class"] = type_
        else:
            self._field = FIELD_MAPPING[self.type]

    def __call__(self, *args, **kwargs):
        nullable = kwargs.pop("nullable", False) or self.nullable

        # This is necessary to allow the following syntax in the model:
        #     field_name: Field[str](max_char=5)
        if not args:
            if self._field is ForeignKeyField and kwargs.get("one_to_one"):
                self._field = OneToOneField
            elif self._field is ForeignKeyField and kwargs.get("many_to_many"):
                self._field = ManyToManyField

            self._kwargs.update(kwargs)
            return self

        kwargs.update(self._kwargs)
        return self._field(*args, python_type=self.type, nullable=nullable, **kwargs)


class BaseField:
    PYTHON_TYPE = None

    def __init__(self, parent, field_name, python_type=None, nullable=False, **kwargs):
        self._parent = parent
        self.name = field_name
        self.python_type = python_type or self.PYTHON_TYPE
        self.nullable = nullable
        self.default = None
        self.is_pk = kwargs.get("primary_key", False)

    def validate(self, value):
        if value is None and not self.nullable and not self.is_pk:
            raise FieldValidationError(f"Field {self.name} cannot be None.")

        if value is not None and not isinstance(value, self.PYTHON_TYPE):
            raise FieldValidationError(
                f"Field {self.name} on model {self._parent.__name__} can only accept type {self.PYTHON_TYPE}, {type(value)} was passed."
            )

    def clean_value(self, value):
        if value is None and self.default:
            value = self.default

        return value

    def build_for_model(self, value):
        return value

    def get_relation_wrapper(self, field_name, value):
        return

    def available_lookups(self):
        return {"equal": Lookup, "in": InLookup}

    def get_query_table_name(self):
        return self._parent.table_name


class StringField(BaseField):
    PYTHON_TYPE = str

    def __init__(self, *args, **kwargs):
        self.max_char = kwargs.pop("max_char")
        super().__init__(*args, **kwargs)


class IntegerField(BaseField):
    PYTHON_TYPE = int


class FloatField(BaseField):
    PYTHON_TYPE = float


class RelationField(BaseField):
    pass


class BaseWrapper:
    pass


class ForeignKeyWrapper(BaseWrapper):
    def __init__(self, foreign_model, id):
        self._foreign_model = foreign_model
        self.id = id

        self._cache = None

    def __call__(self):
        if self.id is None:
            return None

        if self._cache is not None:
            return self._cache

        self._cache = self._foreign_model.get(id=self.id)
        return self._cache


class ForeignKeyField(RelationField):
    PYTHON_TYPE = int

    def __init__(self, parent, field_name, foreign_model_class, **kwargs):
        self._foreign_model = foreign_model_class
        self._model_field_name = field_name
        field_name = self._get_field_name()
        super().__init__(parent, field_name, **kwargs)

        reverse_relation_name = kwargs.get(
            "reverse_name", self._get_default_reverse_relation_name()
        )
        self._set_reverse_relation(reverse_relation_name)

    def _set_reverse_relation(self, reverse_relation_name):
        if hasattr(self._foreign_model, reverse_relation_name):
            raise FieldValidationError(
                f"{reverse_relation_name} already exists on the reverse relation for {self._model_field_name}. "
                "To fix this, set reverse_name on the field definition like so: Field[Model](reverse_name='my_reverse_name')"
            )

        setattr(
            self._foreign_model,
            reverse_relation_name,
            self._get_reverse_relation(self.name),
        )

    def _get_field_name(self):
        return self._model_field_name + "_id"

    def _get_default_reverse_relation_name(self):
        return self._parent.table_name + "_set"

    def _get_reverse_relation(self, field_name):
        return RelationManager(field_name, self._parent, self._parent)

    def clean_value(self, value):
        if value is not None and hasattr(value, "id") and hasattr(value, "save"):
            if value.id is None:
                value.save()

            value = value.id
        return super().clean_value(value)

    def get_relation_wrapper(self, field_name, value):
        return ForeignKeyWrapper(self._foreign_model, value)

    def available_lookups(self):
        model_lookups = super().available_lookups()

        return {**model_lookups, **self._foreign_model.available_lookups()}


class OneToOneWrapper(BaseWrapper):
    def __init__(self, foreign_model):
        self._foreign_model = foreign_model
        self._cache = None
        self.id = None

    def __call__(self):
        if self.id is None:
            return None

        if self._cache is not None:
            return self._cache

        self._cache = self._foreign_model.get(id=self.id)
        return self._cache


class OneToOneField(ForeignKeyField):
    def _get_default_reverse_relation_name(self):
        return self._parent.table_name

    def _get_reverse_relation(self, _field_name):
        return OneToOneWrapper(self._parent)


class ManyToManyField(ForeignKeyField):
    PYTHON_TYPE = Iterable

    def __init__(self, parent, field_name, foreign_model_class, **kwargs):
        nullable = kwargs.get("nullable", False)
        # kwargs["nullable"] = True
        super().__init__(parent, field_name, foreign_model_class, **kwargs)

        self.reverse_name = kwargs.get(
            "reverse_name", self._get_default_reverse_relation_name()
        )

        self._through_model = kwargs.get(
            "through",
            get_through_model(
                parent,
                foreign_model_class,
                self._get_field_name(),
                self.reverse_name,
                nullable,
            ),
        )

        self._mapping = {
            self._get_field_name(): self.reverse_name,
            self.reverse_name: self._get_field_name(),
        }

    def _set_reverse_relation(self, reverse_relation_name):
        return

    def _get_field_name(self):
        return self._model_field_name

    def get_relation_wrapper(self, field_name, value):
        return ManyToManyManager(self._through_model, self._foreign_model)

    def clean_value(self, value):
        if value is None:
            return

        if not isinstance(value, Iterable):
            value = [value]

        return value

    def get_query_table_name(self):
        return self._foreign_model.table_name


FIELD_MAPPING = {str: StringField, int: IntegerField, float: FloatField}



import sqlparse
from ..base import BaseDatabaseSchemaEditor, BaseDatabaseSchemaReader


DEFAULT_VALUE = "default_value"


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    def build_sql_migration(self, diff):
        statements = []

        statements.extend(self._build_delete_statements(diff["delete"]))
        statements.extend(self._build_alter_statements(diff["alter"]))
        statements.extend(self._build_create_statements(diff["create"]))

        sql = sqlparse.format(
            "".join(statements),
            keyword_case="upper",
            identifier_case="lower",
            reindent=True,
        )
        return sql

    def _build_delete_statements(self, diff):
        statements = []

        for table_name in diff:
            statements.append(f"DROP TABLE {table_name};")

        return statements

    def _build_alter_statements(self, diff):
        statements = []

        for table_name, changes in diff.items():
            if not changes["alter"] and not changes["delete"]:
                for row in changes["create"]:
                    statement = f"ALTER TABLE {table_name} ADD "
                    statement += self._format_column(row, add_comma=False)
                    statement += ";"
                    statements.append(statement)

            else:
                old_table_name = f"_{table_name}__old"
                statements.append(
                    f"ALTER TABLE {table_name} RENAME TO {old_table_name};"
                )
                statements.append(
                    self._format_create_table(
                        table_name,
                        (
                            *changes["alter"],
                            *changes["create"],
                            *changes["no_change"],
                        ),
                    )
                )
                statements.append(
                    self._transfer_data(
                        old_table_name,
                        table_name,
                        (*changes["alter"], *changes["no_change"]),
                    )
                )
                statements.append(f"DROP TABLE {old_table_name};")

        return statements

    def _build_create_statements(self, diff):
        statements = []

        for table_name, rows in diff.items():
            statements.append(self._format_create_table(table_name, rows))

        return statements

    def _format_create_table(self, table_name, rows):
        statement = f"CREATE TABLE {table_name} ("
        for i, row in enumerate(rows):
            statement += self._format_column(row, add_comma=(i + 1 != len(rows)))

        statement += ");"
        return statement

    def _format_column(self, row, add_comma=True):
        return (
            f"{row['name']} {row['type']}{' NOT NULL' if row['notnull'] else ''}"
            f"{f' DEFAULT {row[DEFAULT_VALUE]}' if row[DEFAULT_VALUE] is not None else ''}"
            f"{' PRIMARY KEY' if row['is_pk'] else ''}{',' if add_comma else ''}"
        )

    def _transfer_data(self, old_table_name, table_name, rows):
        col_names = [row["name"] for row in rows]
        col_names = ", ".join(col_names)
        statement = (
            f"INSERT INTO {table_name} ({col_names}) "
            f"SELECT {col_names} FROM {old_table_name};"
        )
        return statement


class DatabaseSchemaReader(BaseDatabaseSchemaReader):
    def get_tables(self):
        tables = self._db_client.execute(
            "SELECT name FROM sqlite_schema WHERE type='table' AND name NOT LIKE 'sqlite_%';"
        ).fetchall()
        return [table[0] for table in tables]

    def get_rows(self, table_name):
        rows = self._db_client.execute(f"PRAGMA table_info({table_name});").fetchall()

        formatted_rows = []
        for row in rows:
            formatted_rows.append(
                {
                    "name": row[1],
                    "type": row[2],
                    "notnull": bool(row[3]),
                    "default_value": row[4],
                    "is_pk": bool(row[5]),
                }
            )

        return formatted_rows



from rogue.errors import ValidationError


class NotAKnownFieldException(Exception):
    pass


class DataNotFetchedException(Exception):
    pass


class MissingFieldValueError(Exception):
    pass


class FieldValidationError(ValidationError):
    pass


class ModelValidationError(ValidationError):
    pass



from importlib import import_module
from inspect import ismodule
from rogue.settings import settings


def get_through_model(
    model_has_definition,
    model_inherits_field,
    model_has_definition_reverse_name,
    model_inherits_field_reverse_name,
    nullable=False,
):
    from .base import ModelMeta, Model
    from .fields import Field

    through_name = (
        f"{model_has_definition.__name__}"
        f"{model_has_definition_reverse_name.title().replace('_', '')}"
        f"{model_inherits_field.__name__}"
        f"{model_inherits_field_reverse_name.title().replace('_', '')}Through"
    )

    return ModelMeta(
        through_name,
        (Model,),
        {
            "__annotations__": {
                _get_field_name(model_has_definition): Field[model_has_definition](
                    reverse_name=model_has_definition_reverse_name, nullable=nullable
                ),
                _get_field_name(model_inherits_field): Field[model_inherits_field](
                    reverse_name=model_inherits_field_reverse_name, nullable=nullable
                ),
            }
        },
    )


def _get_field_name(model):
    return model.table_name


def get_all_models():
    # Likely to have a circular import in the future, so we import directly in the function
    from .base import Model

    _import_all_models()

    subclasses = _get_subclasses(Model)
    return subclasses


def _import_all_models():
    # Models have to be imported first to be in the subclasses list
    models_folder = settings.MODELS_FOLDER
    import_module(models_folder)


def _get_subclasses(cls):
    subclasses = []
    for subclass in cls.__subclasses__():
        if subclass.__subclasses__():
            subclasses.extend(_get_subclasses(subclass))
        else:
            subclasses.append(subclass)

    return subclasses



class Lookup:
    comparison = "equal"

    def __init__(self, obj, value, tracking) -> None:
        self.parent = obj
        self.value = value
        self.tracking = tracking


class InLookup(Lookup):
    comparison = "in"



from collections.abc import Iterable

from rogue.backends.sqlite.client import DatabaseClient
from rogue.backends.sqlite.query import QueryBuilder
from rogue.query import Lookup, RelationDescriptor

from .errors import ManagerValidationError


LOOKUP_SEPARATOR = "__"


class Manager:
    def __init__(self, model_class, parent=None):
        self.model_class = model_class
        self._parent = None

        self._client = DatabaseClient(self.model_class.db_name)
        self._query = QueryBuilder(self._client, self.model_class)
        self._cache = None
        self._is_none = False

        self._joins = {}

    def first(self):
        if self._is_none:
            return

        self._base_filtering()

        data = self._build_models(self._query.fetch_one())
        try:
            return data[0]
        except (IndexError, TypeError):
            return

    def all(self):
        return self

    def where(self, not_=False, table_name=None, **where):
        where = self._deconstruct_where(where)
        if where:
            # TODO: Filter the data instead of forcing a refetch
            self._cache = None
            for lookup in where:
                relation_descriptor = (
                    RelationDescriptor(*lookup.tracking)
                    if len(lookup.tracking) > 1
                    else None
                )
                self._query = self._query.where(
                    table_name=table_name or lookup.parent.get_query_table_name(),
                    field=lookup.parent.name,
                    comparison=lookup.comparison,
                    value=lookup.value,
                    not_=not_,
                    relation_descriptor=relation_descriptor,
                )
        return self

    def where_not(self, **where):
        return self.where(not_=True, **where)

    def insert(self, data):
        self.validate_data(data)
        return self._query.insert(data)[0]

    def update(self, pk, data):
        self.validate_data(data)
        return self._query.update(pk, data)[0]

    def delete(self, pk):
        self.where(id=pk)._query.delete()

    def none(self):
        self._is_none = True
        return self

    def validate_data(self, data):
        if data is None:
            raise ManagerValidationError(
                "The data argument must be passed for insert or update."
            )

        model_fields = self.model_class.get_fields()

        for field_name in data:
            if field_name not in self.model_class.get_field_names():
                raise ManagerValidationError(
                    f"{self.model_class.table_name} has no field named {field_name}."
                )
            model_fields[field_name].validate(data[field_name])

    def available_lookups(self):
        return self.model_class.available_lookups()

    def _deconstruct_where(self, where):
        formatted_where = []

        for lookup_str, value in where.items():
            lookup = self._get_lookup_object(lookup_str, value)
            formatted_where.append(lookup)

        return formatted_where

    def _get_lookup_object(self, lookup: str, value):
        available_lookups = self.available_lookups()
        prev_obj = None
        obj = None

        tracking = []
        for item in lookup.split(LOOKUP_SEPARATOR):
            try:
                prev_obj = obj
                obj = available_lookups[item]
                tracking.append(obj)
            except KeyError:
                raise LookupError(
                    f"{item} is not a valid lookup. Options are {', '.join(available_lookups)}."
                )

            if not hasattr(obj, "available_lookups"):
                break

            available_lookups = obj.available_lookups()

        if isinstance(obj, type) and issubclass(obj, Lookup):
            return obj(prev_obj, value, tracking)

        return Lookup(obj, value, tracking)

    def _build_models(self, data):
        models = []
        for row in data:
            row = self._build_relations(row)
            model_class = self.get_returned_model_class()
            models.append(model_class(id_=row.get("id"), parent=self, **row))

        return models

    def _build_relations(self, row):
        for field_name, field in self.model_class.get_related_fields().items():
            if field.name in row:
                row[field_name] = row[field.name]
            elif hasattr(field, "_through_model"):
                relation_manager = field.get_relation_wrapper(field_name, None)
                relation_manager.id = row.get("id")
                row[field_name] = relation_manager
        return row

    def get_returned_model_class(self):
        return self.model_class

    def get_query_table_name(self):
        return self.model_class.table_name

    def __iter__(self):
        return iter(self.all_models)

    @property
    def all_data(self):
        obj = self._base_filtering()

        if self._is_none or self.model_class.get_field_names() == ["id"]:
            return []

        if obj._cache is not None:
            data = obj._cache
        else:
            data = obj._query.fetch_all()
            obj._cache = data

        return data

    def _base_filtering(self):
        return self

    @property
    def all_models(self):
        return self._build_models(self.all_data)

    def __eq__(self, other):
        return self.all_data == other.all_data

    def __len__(self):
        return len(list(self.__iter__()))

    def __contains__(self, other):
        if not isinstance(other, self.model_class):
            return False

        for data in self.all_data:
            if other.id == data["id"]:
                return True

        return False

    def __repr__(self):
        return f"<{self.__class__.__name__} [{', '.join(str(model) for model in self.all_models)}]>"


class RelationManager(Manager):
    def __init__(self, lookup_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lookup_field = lookup_field
        self.id = None

    def _base_filtering(self):
        if self.id is None:
            return self.none()

        return self.where(**{self.lookup_field: self.id})


class ManyToManyManager(Manager):
    def __init__(self, through_model, relation_model):
        super().__init__(through_model)
        self.lookup_field = None
        self.id = None

        self.relation_model = relation_model

    def _base_filtering(self):
        if self.id is None:
            return self.none()

        self._query.model = self.relation_model
        return self.where(table_name=self.get_query_table_name(), id=self.id)

    def add(self, data):
        # TODO: Add a way to insert many with one query
        if isinstance(data, Iterable):
            all_data = []
            for row in data:
                insert_data = self._get_ids_from_row(row)
                all_data.append(self.insert(insert_data))
            return all_data

        insert_data = self._get_ids_from_row(data)
        return self.insert(insert_data)

    def _get_ids_from_row(self, row):
        if not row:
            return {}

        return {row.table_name + "_id": row.id, self.lookup_field: self.id}

    def get_returned_model_class(self):
        return self.relation_model

    def get_query_table_name(self):
        return self.relation_model.table_name



from copy import copy
from typing import Any
import re

from rogue.managers import Manager, RelationManager, ManyToManyManager
from rogue.settings import settings

from .fields import (
    BaseField,
    Field,
    RelationField,
    OneToOneWrapper,
    BaseWrapper,
    ManyToManyField,
)


class ModelMeta(type):
    def __new__(
        cls: type,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwds: Any,
    ):
        instance = super().__new__(cls, name, bases, namespace, **kwds)

        if name == "Model":
            return instance

        instance.table_name = cls._get_table_name(name)
        instance.id = Field[int](instance, "id", primary_key=True)

        __annotations__ = namespace.get("__annotations__") or {}

        for name, obj in __annotations__.items():
            if obj is Field:
                raise TypeError(
                    "Field must be used with type hinting. For example, call it like so: Field[str]."
                )
            if isinstance(obj, Field):
                obj = obj(instance, name)
            if isinstance(obj, BaseField) and name in namespace:
                obj.default = namespace[name]
            setattr(instance, name, obj)

        return instance

    @classmethod
    def _get_table_name(cls, name):
        return "_".join(re.sub(r"([A-Z])", r" \1", name).split()).lower()


class Model(metaclass=ModelMeta):
    db_name = settings.DATABASE_NAME

    def __init__(self, id_=None, parent=None, **kwargs):
        self._parent = parent

        self._foreign_relations = {}

        self._set_related_managers()
        self._set_related_managers_id(id_)

        for field_name, field in self.get_model_fields().items():
            value = field.clean_value(kwargs.pop(field_name, None))
            value = field.build_for_model(value)
            field.validate(value)

            if field.name:
                setattr(self, field.name, value)

            foreign_relations = field.get_relation_wrapper(field.name, value)
            if isinstance(foreign_relations, ManyToManyManager):
                foreign_relations.id = id_
                foreign_relations.lookup_field = self.table_name + "_id"
            if foreign_relations is not None:
                self._foreign_relations[field_name] = foreign_relations

        for attr, value in kwargs.items():
            setattr(self, attr, value)

        self.__values_last_save = self.field_values

    def _set_related_managers(self):
        for field, manager in self.get_class_related_managers().items():
            setattr(self, field, copy(manager))

    def _set_related_managers_id(self, id):
        for manager in self.get_related_managers().values():
            manager.id = id

        for manager in self.get_many_managers().values():
            manager.id = id

    @classmethod
    def _get_new_manager(cls):
        return Manager(cls)

    def save(self):
        created = False
        manager = self._get_new_manager()
        field_values = self.field_values

        if self.id is None:
            created = True
            del field_values["id"]
            new_values = manager.insert(field_values)
            self.id = new_values["id"]
            self.__values_last_save["id"] = self.id
            self._set_related_managers_id(self.id)
        else:
            manager.update(self.id, self.get_changed_fields())

        if created:
            for field_name, field in self.get_many_managers().items():
                if field_name in self.__dict__:
                    value = self.__dict__.get(field_name)
                    if value is not None:
                        field.add(value)

        if not created:
            for field_name, value in field_values.items():
                setattr(self, field_name, value)
                self.__values_last_save[field_name] = value

        return created

    def delete(self):
        self._get_new_manager().delete(self.id)
        self.id = None

    @classmethod
    def get(cls, **kwargs):
        return cls._get_new_manager().where(**kwargs).first()

    @classmethod
    def where(cls, **kwargs):
        return cls._get_new_manager().where(**kwargs)

    @classmethod
    def where_not(cls, **kwargs):
        return cls._get_new_manager().where_not(**kwargs)

    @classmethod
    def all(cls):
        return cls._get_new_manager().all()

    @classmethod
    def none(cls):
        return cls._get_new_manager().none()

    @classmethod
    def get_fields(cls):
        return {
            field.name: field
            for field in cls.__dict__.values()
            if isinstance(field, BaseField)
            and field.name
            and not isinstance(field, ManyToManyField)
        }

    @classmethod
    def get_model_fields(cls):
        return {
            field_name: field
            for field_name, field in cls.__dict__.items()
            if isinstance(field, BaseField)
        }

    @classmethod
    def available_lookups(cls):
        available_lookups = {}

        for field_name, field in cls.get_model_fields().items():
            available_lookups[field_name] = field

            if field.name != field_name:
                available_lookups[field.name] = field

        return available_lookups

    @classmethod
    def get_field_names(cls):
        return list(cls.get_fields())

    @classmethod
    def get_related_fields(cls):
        return {
            field_name: field
            for field_name, field in cls.get_model_fields().items()
            if isinstance(field, RelationField)
        }

    @classmethod
    def get_class_related_managers(cls):
        return {
            field_name: field
            for field_name, field in cls.__dict__.items()
            if isinstance(field, (RelationManager, OneToOneWrapper))
        }

    def get_related_managers(self):
        return {
            field_name: field
            for field_name, field in self.__dict__.items()
            if isinstance(field, (RelationManager, OneToOneWrapper))
        }

    def get_many_managers(self):
        return {
            field_name: field
            for field_name, field in self.__dict__.get("_foreign_relations", {}).items()
            if isinstance(field, ManyToManyManager)
        }

    @property
    def field_values(self):
        return {
            field.name: getattr(self, field_name)
            for field_name, field in self.get_fields().items()
            if field.name
        }

    def get_changed_fields(self):
        changed_fields = {}

        for field, value in self.field_values.items():
            if value != self.__values_last_save.get(field):
                changed_fields[field] = value

        return changed_fields

    def __setattr__(self, attr, value):
        fields = self.get_fields()

        if attr in fields:
            fields[attr].validate(value)

        super().__setattr__(attr, value)

    def __getattribute__(self, name: str):
        attribute = super().__getattribute__(name)

        if name != "_foreign_relations" and name in getattr(
            self, "_foreign_relations", []
        ):
            attribute = self._foreign_relations[name]

        if isinstance(attribute, BaseWrapper):
            return attribute()

        return attribute

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, self.__class__) and obj.id == self.id

    def __repr__(self) -> str:
        fields = [
            f"{str(field)}={str(value)}" for field, value in self.field_values.items()
        ]
        return f"<{self.__class__.__name__} {', '.join(fields)}>"



from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any

from rogue.settings import settings

from .errors import OperationalError, InvalidComparisonError


class BaseDatabaseClient(metaclass=ABCMeta):
    _instances = {}
    _db_name = None

    def __new__(cls, db_name=settings.DATABASE_NAME):
        if db_name not in cls._instances:
            cls._instances[db_name] = super().__new__(cls)

        return cls._instances[db_name]

    def __init__(self, db_name=settings.DATABASE_NAME):
        self._db_name = db_name

        self._connection = None

    @abstractmethod
    def get_connection(self):  # pragma: no cover
        pass

    @abstractmethod
    def execute(self, statement, *args, **kwargs):  # pragma: no cover
        pass

    @abstractmethod
    def close(self):  # pragma: no cover
        pass


@dataclass
class WhereStatement:
    table_name: str
    field: str
    comparison: str
    value: Any
    relation_descriptor: Any


class BaseQueryBuilder(metaclass=ABCMeta):
    SELECT = "SELECT"
    UPDATE = "UPDATE"
    INSERT = "INSERT INTO"
    DELETE = "DELETE"

    FROM = "FROM"
    WHERE = "WHERE"
    AND = "AND"
    VALUES = "VALUES"
    INNER_JOIN = "INNER JOIN"
    ON = "ON"

    EQUAL = "equal"
    IN = "in"

    COMPARISON_MAPPING = {
        EQUAL: "=",
        IN: "IN",
    }
    NOT_COMPARISON_MAPPING = {
        EQUAL: "!=",
        IN: "NOT IN",
    }

    COMPARISON_DEFAULT = EQUAL

    def __init__(self, client, model):
        self.client = client
        self.model = model

        self.where_statements = []

    @property
    def table_name(self):
        return self.model.table_name

    @property
    def fields(self):
        return self.model.get_fields()

    def where(
        self,
        *,
        table_name,
        field,
        comparison,
        value,
        not_: bool = False,
        relation_descriptor=None,
    ):
        comparison_mapping = (
            self.NOT_COMPARISON_MAPPING if not_ else self.COMPARISON_MAPPING
        )

        if not comparison:
            comparison = self.COMPARISON_DEFAULT

        try:
            comparison = comparison_mapping[comparison]
        except KeyError:
            raise InvalidComparisonError(
                f"{comparison} is not a valid comparison operator."
            )

        self.where_statements.append(
            WhereStatement(
                table_name=table_name,
                field=field,
                comparison=comparison,
                value=value,
                relation_descriptor=relation_descriptor,
            )
        )
        return self

    @abstractmethod
    def fetch_one(self):  # pragma: no cover
        pass

    @abstractmethod
    def fetch_all(self):  # pragma: no cover
        pass

    @abstractmethod
    def insert(self, data):  # pragma: no cover
        pass

    @abstractmethod
    def update(self, pk, data):  # pragma: no cover
        pass

    @abstractmethod
    def delete(self, pk):  # pragma: no cover
        pass

    def _format_input_row(self, headers, data):
        headers = list(data)
        formatted_data = [data[header] for header in headers]
        return formatted_data

    def _format_input_data(self, data):
        if isinstance(data, dict):
            headers = list(data)
            formatted_data = self._format_input_row(headers, data)

        return headers, formatted_data

    def _format_output_data(self, data):
        assert isinstance(data, Iterable), "Field data must be an iterable."

        formatted_data = []

        if not data or data[0] is None:
            return formatted_data

        for row in data:
            formatted_data.append(dict(zip(self.fields.keys(), row)))

        return formatted_data

    @abstractmethod
    def _build_select(self):  # pragma: no cover
        pass

    @abstractmethod
    def _build_insert(self, data):  # pragma: no cover
        pass

    @abstractmethod
    def _build_update(self, pk, data):  # pragma: no cover
        pass

    @abstractmethod
    def _build_delete(self, pk):  # pragma: no cover
        pass

    @abstractmethod
    def _build_where(self):  # pragma: no cover
        pass

    def _validate_data(self, data):
        assert data is not None, "Cannot insert without data."


class BaseDatabaseSchemaEditor(metaclass=ABCMeta):
    def __init__(self, db_client: BaseDatabaseClient) -> None:
        self._db_client = db_client


class BaseDatabaseSchemaReader(metaclass=ABCMeta):
    def __init__(self, db_client: BaseDatabaseClient) -> None:
        self._db_client = db_client

    @abstractmethod
    def get_tables(self):  # pragma: no cover
        pass

    @abstractmethod
    def get_rows(self, table_name):  # pragma: no cover
        pass
